import json
import logging
import sys
import traceback
from argparse import ArgumentParser, Namespace
from typing import Optional

from prettytable import PrettyTable

import pysonrpc

log = logging.getLogger(__name__)


def command_list(cli: pysonrpc.JsonRpcEndpoint, args: Namespace):
    if args.filter:
        met_list = [met for met in cli.methods.values() if args.filter in met.fullname]
    else:
        met_list = cli.methods.values()
    if args.raw:
        fullprops = {met.fullname: met.properties for met in met_list}
        print(json.dumps(fullprops, indent=2))
    else:
        cols = ["Method", "Parameters"] if args.short else ["Method", "Parameters", "Description"]
        tab = PrettyTable(cols)
        tab.align = "l"
        for met in met_list:
            row = [met.fullname, met.param_list()]
            if not args.short:
                row.append(met.description)
            tab.add_row(row)
        print(tab)


def command_run(cli: pysonrpc.JsonRpcEndpoint, args: Namespace):
    params = json.loads(args.params)
    result = cli.run_method(args.method, **params, raw=args.raw)
    print(json.dumps(result, indent=2))


def _setup_logging(level: int = logging.INFO, filename: Optional[str] = None) -> None:
    """Configure standard logging."""
    logging.basicConfig(
        format="%(asctime)s %(filename)s:%(lineno)s %(levelname)s: %(message)s", level=level, filename=filename
    )


def _parse_args() -> Namespace:
    parser = ArgumentParser(description="RPC client")

    parser.add_argument("--version", "-v", help="Display version", default=False, action="store_true")
    parser.add_argument("--url", "-r", help="Host url, e.g 'http://192.168.0.1:8080'", required=True)
    parser.add_argument("--user", "-u", help="username if using basic authentication", default=None)
    parser.add_argument("--password", "-p", help="Password if using basic authentication", default=None)
    parser.add_argument("--debug", "-d", default=False, action="store_true", help="Enable debug logging")
    parser.add_argument(
        "--schema-discover",
        "-a",
        default=None,
        action="store_true",
        help="Auto discover rpc methods schema at the endpoint url",
    )
    parser.add_argument(
        "--schema-discover-path",
        "-ap",
        default=None,
        help="Auto discover rpc methods schema from this path (relative to the endpoint url)",
    )
    parser.add_argument(
        "--schema-discover-method",
        "-am",
        default=None,
        help="Auto discover rpc methods schema by calling this json rpc method",
    )
    parser.add_argument("--method-file", "-f", help="Discover methods from given json file", default=None)
    # json_file
    # auto_discover
    subparsers = parser.add_subparsers(help="commands", dest="command", required=True)

    # List command
    list_parser = subparsers.add_parser("list", help="List available methods")
    list_parser.add_argument("--raw", "-j", default=False, action="store_true", help="Display full description")
    list_parser.add_argument("--short", "-s", default=False, action="store_true", help="Only display name and params")
    list_parser.add_argument("--filter", "-f", default=None, help="Filter RPC methods names to print")
    list_parser.set_defaults(func=command_list)

    # Run command
    run_parser = subparsers.add_parser("run", help="Execute a method")
    run_parser.add_argument("--method", "-m", default=None, help="RPC method to execute", required=True)
    run_parser.add_argument(
        "--params", "-p", default="{}", help="Optional parameters for the method as json, e.g: '{id:1, name:\"test\"}'"
    )
    run_parser.add_argument("--raw", "-j", default=False, action="store_true", help="Raw json response")
    run_parser.set_defaults(func=command_run)

    args = parser.parse_args()
    args.log_level = logging.DEBUG if args.debug else logging.INFO

    return args


def main() -> None:
    """Main entry point."""

    args = _parse_args()
    _setup_logging(level=args.log_level)

    if args.version:
        print(f"pysonrpc version {pysonrpc.__version__}", file=sys.stderr)

    try:
        cli = pysonrpc.JsonRpcEndpoint(
            args.url,
            user=args.user,
            password=args.password,
            auto_detect=args.schema_discover,
            schema_path=args.schema_discover_path,
            schema_method=args.schema_discover_method,
            json_file=args.method_file,
        )

        if args.func:
            args.func(cli, args)
        else:
            raise pysonrpc.JsonRpcClientError(f"No processing defined for command {args.command}")

        sys.exit(0)
    except pysonrpc.JsonRpcClientError as e:
        print(f"Client error: {e}")
        if args.debug:
            traceback.print_exc()
        sys.exit(1)
    except Exception as e:
        print(f"Unexpected error: {e}")
        if args.debug:
            traceback.print_exc()
        sys.exit(2)



import json
import logging
import uuid
from typing import Any, Dict, List, Optional, Union

import requests

log = logging.getLogger(__name__)


class JsonRpcError(Exception):
    pass


class JsonRpcClientError(JsonRpcError):
    pass


class JsonRpcServerError(JsonRpcError):
    pass


class JsonRpcClient:
    """Implementation of a json rpc client."""

    # RPC message config
    JSONRPC_VERSION = "2.0"
    JSONRPC_CONTENT = "application/json"

    # Defined error codes
    JSONRPC_PARSE_ERROR = -32700
    JSONRPC_INVALID_REQUEST = -32600
    JSONRPC_METHOD_NOT_FOUND = -32601
    JSONRPC_INVALID_PARAMS = -32602
    JSONRPC_INTERNAL_ERROR = -32603
    # SERVER_ERROR = -32000 to -32099

    # jsonrpc dictionnary keys
    JSONRPC_KEY = "jsonrpc"
    JSONRPC_KEY_ID = "id"
    JSONRPC_KEY_REQ_METHOD = "method"
    JSONRPC_KEY_REQ_PARAMS = "params"
    JSONRPC_KEY_RESP_RESULT = "result"
    JSONRPC_KEY_RESP_ERROR = "error"
    JSONRPC_KEY_RESP_ERROR_CODE = "code"
    JSONRPC_KEY_RESP_ERROR_MSG = "message"
    JSONRPC_KEY_RESP_ERROR_DATA = "data"

    def __init__(self, url, user: Optional[str] = None, password: Optional[str] = None) -> None:
        self._url = url
        self._auth = self._build_credentials(user, password)

    def _build_credentials(self, user: Optional[str], password: Optional[str]) -> Optional[Any]:
        """Build http basic auth credentials per default."""
        if user and password:
            return requests.auth.HTTPBasicAuth(user, password)
        return None

    def _random_id(self) -> str:
        return uuid.uuid4().hex

    def _build_jsonrpc_payload(
        self, method: str, params: Dict[str, Any] = {}, req_id: Optional[Union[int, str]] = None
    ) -> Dict[str, Any]:
        """Create JSON-RPC payload."""
        return {
            self.JSONRPC_KEY: self.JSONRPC_VERSION,
            self.JSONRPC_KEY_REQ_METHOD: method,
            self.JSONRPC_KEY_REQ_PARAMS: params,
            self.JSONRPC_KEY_ID: req_id or self._random_id(),
        }

    def _parse_response(self, response: requests.Response, raw: bool = True) -> Dict[str, Any]:
        """Extract json result from response upon success or raise an exception."""
        if response is not None:
            if response.status_code == 200:
                try:
                    raw_json = response.json()
                    print(f"decoded result {raw_json}")
                    return raw_json if raw else self.jsonrpc_result(raw_json)
                except json.JSONDecodeError as e:
                    raise JsonRpcServerError(f"Invalid json response: {response.text}") from e

            raise JsonRpcServerError(f"Couldn't get data response code {response.status_code}: {response.reason}")
        raise JsonRpcServerError(f"Couldn't get response from server: {response}")

    def get(self, path: Optional[str] = None, headers: Dict[str, str] = {}) -> Dict[str, Any]:
        """Send a get requests to the server."""
        headers.update(
            {
                "Content-Type": self.JSONRPC_CONTENT,
            }
        )

        # Make the JSON-RPC request using the requests library
        url = f"{self._url}/{path}" if path else self._url
        log.debug(f"JSON RPC get to {url}")
        try:
            response = requests.get(url, headers=headers, auth=self._auth)
        except Exception as e:
            raise JsonRpcClientError(f"Request error: {e}") from e

        return self._parse_response(response)

    def request(
        self,
        method,
        params={},
        req_id: Optional[Union[int, str]] = None,
        headers: Dict[str, str] = {},
        raw: bool = True,
    ) -> Dict[str, Any]:
        """Sends a json rpc request, with optional headers and id and return the json result or the raw response."""
        payload = self._build_jsonrpc_payload(method, params, req_id)
        headers.update(
            {
                "Content-Type": self.JSONRPC_CONTENT,
            }
        )

        # Make the JSON-RPC request using the requests library
        log.debug(f"JSON RPC request to {self._url}: {payload}")
        try:
            response = requests.post(self._url, json=payload, headers=headers, auth=self._auth)
        except Exception as e:
            raise JsonRpcClientError(f"Request error: {e}") from e

        return self._parse_response(response, raw)

    def jsonrpc_error(
        self, error: int, message: str, data: Optional[Any] = None, req_id: Optional[int] = None
    ) -> Dict[str, Any]:
        error_message = {
            self.JSONRPC_KEY: self.JSONRPC_VERSION,
            self.JSONRPC_KEY_ID: req_id,
            self.JSONRPC_KEY_RESP_ERROR: {
                self.JSONRPC_KEY_RESP_ERROR_CODE: error,
                self.JSONRPC_KEY_RESP_ERROR_MSG: message,
            },
        }
        if data:
            error_message[self.JSONRPC_KEY_RESP_ERROR][self.JSONRPC_KEY_RESP_ERROR_DATA] = data  # type: ignore
        return error_message

    def jsonrpc_result(self, response: Dict[str, Any]) -> Dict[str, Any]:
        if self.JSONRPC_KEY_RESP_RESULT in response:
            return response[self.JSONRPC_KEY_RESP_RESULT]
        elif self.JSONRPC_KEY_RESP_ERROR in response:
            raise JsonRpcServerError(response[self.JSONRPC_KEY_RESP_ERROR].get(self.JSONRPC_KEY_RESP_ERROR_MSG))
        else:
            raise JsonRpcServerError(f"Invalid json rpc response: {response}")


class MethodContainer:
    """Base class to provides a method list and attributes for those method names."""

    _child_methods: Dict[str, Any] = {}

    @property
    def child_methods(self) -> Dict[str, Any]:
        return self._child_methods

    def _execute(self) -> Any:
        """Returns the element to assign to the class attribute with named this method."""

        return self

    def __getattr__(self, name: str) -> Any:
        """If the request attribute doesn't exist, check if it's a method name and get its element if it is."""
        if name in self.child_methods:
            return self.child_methods[name]._execute()
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")


class Method(MethodContainer):
    """Wraps a jsonrpc method description, information, and execution."""

    # JSON Schema properties
    PROP_DESC = "description"
    PROP_PARAMS = "params"
    PROP_RETURNS = "returns"
    PROP_PARAM_NAME = "name"
    NAMESPACE_SEP = "."

    def __init__(self, name, properties={}, exec: bool = True, client: Optional[JsonRpcClient] = None) -> None:
        self._fullname = name
        self._parents = name.split(self.NAMESPACE_SEP)
        self._name = self._parents.pop()
        self._properties = properties or {}
        self._exec = exec
        self._client = client

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}:{self.name}{self.param_list()}"

    def _execute(self) -> Any:
        """Overrides affectation to return itself if only a namespace, or the execution if a actual method."""
        return self.run if self._exec else self

    def run(self, *args, raw=True, **kwargs) -> Dict[str, Any]:
        """Executes this method."""
        if self._client:
            return self._client.request(method=self._fullname, params=kwargs, raw=raw)
        return {}

    def param_list(self) -> List[str]:
        return [param.get(self.PROP_PARAM_NAME) for param in self.params]

    @property
    def fullname(self):
        return self._fullname

    @property
    def parents(self):
        return self._parents

    @property
    def name(self):
        return self._name

    @property
    def properties(self):
        return self._properties

    @property
    def description(self):
        return self._properties.get(self.PROP_DESC)

    @property
    def params(self):
        return self._properties.get(self.PROP_PARAMS, {})

    @property
    def returns(self):
        return self._properties.get(self.PROP_RETURNS)


class JsonRpcEndpoint(MethodContainer):
    """Create a jsonrpc endpoint with methods dynamically defined based on the json schema definition,
    or a list of manually created ones.
    """

    def __init__(
        self,
        url,
        user: Optional[str] = None,
        password: Optional[str] = None,
        json_file: Optional[str] = None,
        schema_path: Optional[str] = None,
        schema_method: Optional[str] = None,
        schema: Optional[Dict[str, Any]] = None,
        auto_detect: Optional[bool] = False,
    ) -> None:
        # Create rpc client
        self.client = JsonRpcClient(url, user, password)
        self._methods: Dict[str, Any] = {}

        # Load methods definition from all defined source: Manual, dict, file, urlx
        methods_list = []
        if json_file:
            methods_list.extend(self._methods_from_file(json_file))
        if schema_method or schema_path or auto_detect:
            methods_list.extend(self._methods_from_url(path=schema_path, method=schema_method))
        if schema:
            methods_list.extend(self._methods_from_dict(schema))

        # Create method hierarachy
        self._add_methods(methods_list)

    @property
    def methods(self):
        return self._methods

    def _add_methods(self, methods_list):
        """From the method list, build direct access list and tree:
        - populate a dict for direct access based on fullname,
        - build a tree based on namespace with leaf being the executable methods
        """

        for method in methods_list:
            parent_node = self.child_methods
            # Create all parent nodes
            for ns in method.parents:
                node = parent_node.setdefault(ns, Method(ns, exec=False))
                parent_node = node.child_methods
            # Add actual method
            parent_node[method.name] = method
            self._methods[method.fullname] = method

    def _methods_from_dict(self, json_schema: Optional[Dict[str, Any]]) -> List[Method]:
        methods = []

        if json_schema and "methods" in json_schema:
            for method_name, method_props in json_schema["methods"].items():
                methods.append(Method(method_name, method_props, client=self.client))
        return methods

    def _methods_from_file(self, json_file: str) -> List[Method]:
        with open(json_file, "r") as fp:
            return self._methods_from_dict(json.load(fp))

    def _methods_from_url(self, path: Optional[str], method: Optional[str]) -> List[Method]:
        if method:
            json_schema = self.client.request(method=method, raw=False)
        else:
            json_schema = self.client.get(path)
        return self._methods_from_dict(json_schema)

    def run_method(self, method, *args, raw: bool = True, **kwargs) -> Dict[str, Any]:
        if self.client:
            return self.client.request(method=method, params=kwargs, raw=raw)
        return {}



import pytest
import json
# from copy import deepcopy
from unittest.mock import ANY, call, patch, Mock
from pysonrpc.jsonrpc import JsonRpcClient, JsonRpcEndpoint, JsonRpcClientError, JsonRpcServerError, Method


# @pytest.fixture
# def cli():
#     return RadarrCli(TEST_HOST, TEST_APIKEY)

TEST_URL="http://127.0.0.1:8080/path"
TEST_HEADERS={'Content-Type': 'application/json'}
TEST_METH_PATH="jsonrpc"
TEST_METH_NAME="introspect"
TEST_METH_FILE="test/methods.json"

TEST_METH_LIST_1 ={
	"methods": {
		"some.method": {
			"description": "some method description",
			"params": [
				{ "name": "param1", "required": True, "type": "string" },
				{ "name": "param3", "required": True, "type": "int" },
				{ "$ref": "some.type", "name": "param2" }
			],
			"returns": {
				"properties": {
					"addon": { "$ref": "Addon.Details", "required": True },
					"limits": { "$ref": "List.LimitsReturned", "required": True }
				},
				"type": "object"
			},
			"type": "method"
		},
		"some.method2": {
			"description": "some method2 description",
			"params": [
			],
			"type": "method"
		},
		"some.method3": {
		},
    }
}

TEST_METH_LIST_2 ={
	"methods": {
		"some2.method": {
			"description": "some method description",
			"params": [
				{ "name": "param1", "required": True, "type": "string" },
				{ "name": "param3", "required": True, "type": "int" },
				{ "$ref": "some.type", "name": "param2" }
			],
			"returns": {
				"properties": {
					"addon": { "$ref": "Addon.Details", "required": True },
					"limits": { "$ref": "List.LimitsReturned", "required": True }
				},
				"type": "object"
			},
			"type": "method"
		},
		"some2.method2": {
			"description": "some method2 description",
			"params": [
			],
			"type": "method"
		},
		"some2.method3": {
		},
    }
}
TEST_METH_NLIST_1 = list(TEST_METH_LIST_1['methods'].keys())
TEST_METH_NLIST_2 = list(TEST_METH_LIST_2['methods'].keys())
TEST_METH_NLIST_3 = ["Some3.Method1", "Some3.Method2"]


def mock_response(code, data_dict):
    resp = Mock()
    resp.status_code = code
    resp.json.side_effect = data_dict
    return resp

def mock_endpoint(mock_post, mock_get, url=None, **kwargs):
    return JsonRpcEndpoint(
        url or TEST_URL,
        user=kwargs.get("user"),
        password=kwargs.get("password"),
        schema=kwargs.get("schema"),
        auto_detect=kwargs.get("auto_detect", False),
        schema_path=kwargs.get("schema_path"),
        schema_method=kwargs.get("schema_method"),
        json_file=kwargs.get("json_file"),
    )

@patch("pysonrpc.jsonrpc.requests.get")
@patch("pysonrpc.jsonrpc.requests.post")
def test_client_no_auth(mock_post, mock_get):
    mock_get.return_value = mock_response(200, ["some response"])
    cli = mock_endpoint(mock_post, mock_get)
    assert cli.methods == {}
    assert cli.client.get() == "some response"
    mock_get.assert_called_with(TEST_URL, headers=TEST_HEADERS, auth=None)


@patch("pysonrpc.jsonrpc.requests.get")
@patch("pysonrpc.jsonrpc.requests.post")
def test_client_auth(mock_post, mock_get):
    mock_get.return_value = mock_response(200, ["some response"])     
    cli = mock_endpoint(mock_post, mock_get, user="user", password="pass")
    assert cli.client.get() == "some response"
    mock_get.assert_called_with(TEST_URL, headers=TEST_HEADERS, auth=cli.client._auth)
    assert cli.client._auth.username == "user"
    assert cli.client._auth.password == "pass"


@patch("pysonrpc.jsonrpc.requests.get")
@patch("pysonrpc.jsonrpc.requests.post")
def test_client_get_error(mock_post, mock_get):
    mock_get.side_effect = Exception("error")
    cli = mock_endpoint(mock_post, mock_get)
    with pytest.raises(JsonRpcClientError):
        cli.client.get()


@pytest.mark.parametrize(
    "auto, path, method, file, expect", [
    (True, None, None, None, TEST_METH_LIST_1['methods'].keys()),
    (False, TEST_METH_PATH, None, None, TEST_METH_NLIST_1),
    (False, None, TEST_METH_NAME, None, TEST_METH_NLIST_2),
    (False, None, None, TEST_METH_FILE, TEST_METH_NLIST_3),
    (True, TEST_METH_PATH, TEST_METH_NAME, TEST_METH_FILE, TEST_METH_NLIST_2+TEST_METH_NLIST_3),
    (True, TEST_METH_PATH, None, TEST_METH_FILE, TEST_METH_NLIST_1+TEST_METH_NLIST_3),
])
@patch("pysonrpc.jsonrpc.requests.get")
@patch("pysonrpc.jsonrpc.requests.post")
def test_client_schema(mock_post, mock_get, auto, path, method, file, expect):
    mock_get.return_value = mock_response(200, [TEST_METH_LIST_1])
    mock_post.return_value = mock_response(200, [{ JsonRpcClient.JSONRPC_KEY_RESP_RESULT: TEST_METH_LIST_2}])
    cli = mock_endpoint(mock_post, mock_get, auto_detect=auto, schema_path=path, schema_method=method, json_file=file)

    for key in expect:
        assert key in cli.methods.keys()
    assert len(expect) == len(cli.methods)

@patch("pysonrpc.jsonrpc.requests.get")
@patch("pysonrpc.jsonrpc.requests.post")
def test_client_attribute_method(mock_post, mock_get):
    methods = {
	    "methods": {
            "some.cool.thing": {},
            "some.bad.withparam": {"params": [{ "name": "param1"},{ "name": "param2"}]},
            "some.cool.alsowithparams": {"params": [{ "name": "param4"},{ "name": "param5"}]},
        }
    }
    pl = { JsonRpcClient.JSONRPC_KEY_RESP_RESULT: "data"}
    mock_post.return_value = mock_response(200, [pl, pl, pl, pl])
    cli = mock_endpoint(mock_post, mock_get, schema=methods)

    assert cli.some.cool.thing() == {"result": "data"}
    assert cli.some.cool.thing(raw = False) == "data"
    assert cli.some.bad.withparam(param1="a", param2="b") == {"result": "data"}
    assert cli.some.cool.alsowithparams(param4="c", param5="d", raw = False) == "data"

    pl1 = { "jsonrpc": "2.0", "method": "some.cool.thing", "params": {}, "id": ANY}
    pl2 = { "jsonrpc": "2.0", "method": "some.bad.withparam", "params": {"param1": "a", "param2": "b"}, "id": ANY}
    pl3 = { "jsonrpc": "2.0", "method": "some.cool.alsowithparams", "params": {"param4": "c", "param5": "d"}, "id": ANY}
    calls = [
        call(TEST_URL, headers=TEST_HEADERS, json=pl1, auth=None),
        call(TEST_URL, headers=TEST_HEADERS, json=pl2, auth=None),
        call(TEST_URL, headers=TEST_HEADERS, json=pl3, auth=None),
    ]
    mock_post.assert_has_calls(calls, any_order=True)

    with pytest.raises(AttributeError):
        cli.some.notcool.thing()


@pytest.mark.parametrize(
    "sideffect, exc", [
    (Exception(), JsonRpcClientError),
    ([None], JsonRpcServerError),
    ([mock_response(200, json.JSONDecodeError("ee", "ee", 21))], JsonRpcServerError),
])
@patch("pysonrpc.jsonrpc.requests.get")
@patch("pysonrpc.jsonrpc.requests.post")
def test_client_method_error(mock_post, mock_get, sideffect, exc):
    pl = { JsonRpcClient.JSONRPC_KEY_RESP_RESULT: "data"}
    # mock_post.side_effect = Exception() // JsonRpcClientError
    # mock_post.side_effect = [None] // JsonRpcServerError 
    # mock_post.return_value = mock_response(200, json.JSONDecodeError("ee", "ee", 21))
    mock_post.side_effect = sideffect
    cli = mock_endpoint(mock_post, mock_get)
    with pytest.raises(exc):
        cli.run_method("method")


@patch("pysonrpc.jsonrpc.requests.get")
@patch("pysonrpc.jsonrpc.requests.post")
def test_client_run_method(mock_post, mock_get):
    pl = { JsonRpcClient.JSONRPC_KEY_RESP_RESULT: "data"}
    mock_post.return_value = mock_response(200, [pl, pl, pl, pl])
    cli = mock_endpoint(mock_post, mock_get)

    assert cli.run_method("method", raw = True) == {"result": "data"}
    assert cli.run_method("some.cool.thing", raw = False) == "data"
    assert cli.run_method("some.bad.withparam", param1="a", param2="b") == {"result": "data"}

    pl1 = { "jsonrpc": "2.0", "method": "method", "params": {}, "id": ANY}
    pl3 = { "jsonrpc": "2.0", "method": "some.cool.thing", "params": {}, "id": ANY}
    pl2 = { "jsonrpc": "2.0", "method": "some.bad.withparam", "params": {"param1": "a", "param2": "b"}, "id": ANY}
    calls = [
        call(TEST_URL, headers=TEST_HEADERS, json=pl1, auth=None),
        call(TEST_URL, headers=TEST_HEADERS, json=pl2, auth=None),
        call(TEST_URL, headers=TEST_HEADERS, json=pl3, auth=None),
    ]
    mock_post.assert_has_calls(calls, any_order=True)
